/*+M*************************************************************************
 * Promise SATAII S150 SX8 device driver for Linux.
 *
 * Copyright (c) 2001  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2001 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *-M************************************************************************
 */
 
/*
 * Description   : safte.c - The codes to achieve saf-te
 */

#if defined(__KERNEL__)
#include <stdarg.h>
#include <asm/io.h>
#include <asm/string.h>
#endif

#include "musafte.h"
#include "smbus.h"
#include "sbox.h"
#include "safteapi.h"
#include "tools.h"

#if !defined(PCHAR)
#define	PCHAR	char *
#endif

#if !defined(NULL)
#define	NULL	((void *)0)
#endif

/*
 * A global varable to store the address of context of sfate
 */
Safte_Context_t  SafteContext[MAX_CARDS] =  {{0}};
pSafte_Context_t pSafteContext[MAX_CARDS] = {NULL, NULL, NULL, NULL};


/* internal function in Safte module */
static int init_saftebox(pSafte_Context_t, int);
static int Safte_GetConf(Enclosure_Conf *, int);

static int Safte_GetEnclosureStatus(char *, int);
static int Safte_GetDevstat(char *, int);
static int Safte_SetDevStat(char *, int);
static int Safte_SetDevOp(char *, int);
static int Safte_SetGlobalCmd(char *, int);


/*
 *
 * The function description:
 *	This function initialize safte module context data, and do some init
 * action on smbus
 *
 * Input parameters:
 *	U32 baddr:	SMBus register base address
 *
 * Output:
 *	If success, pSafteContext will point to the address of buffer holding 
 * safte context, smbus handle will be set to route to firmware,  and smbus will 
 * be initialized, and return value 0
 *   	If failed, psc will be null, and return value will be -1
 *
 */
int InitSafte(u8 *baddr, int cardno)
{
	int ret;

	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] != NULL)
		return(-1);
	
	pSafteContext[cardno] = (pSafte_Context_t)&SafteContext[cardno];
		
	/* Clear memory */
	memset(pSafteContext[cardno], 0, sizeof(Safte_Context_t));
	
	pSafteContext[cardno]->regbaddr = baddr;

	/* Init smbus */
	ret = init_saftebox(pSafteContext[cardno], cardno);

	if(ret < 0)
	{	
		pSafteContext[cardno] = NULL;
		return(-1);
	}

	return(ret);
}


/*
 *
 * The function description:
 *		This function do all Safte module's actions
 *
 * Input Parameters:
 *	 U32 cmd:  SCSI SAFTE R/W command that should be
 *				GET_ENC_CONF,	GET_ENC_STAT,
 *				GET_ENC_USAGE,	GET_DEV_INSERT,
 *				GET_DEV_STAT,
 *				SET_DEV_STAT,	SET_BUS_ID,
 *				SET_DEV_OP,	SET_FAN_SPEED,
 *				SET_POWER_STAT,	SET_GLOB_CMD
 *
 *	 void *data:	The input data if cmd is write command
 *					else the buffer address to hold the return data
 *
 * Output:
 *	  If this function success, return value 0, and 
 *	  pSafte_Context_t psc will hold some safte module's status if cmd is 
 *        write cmd. if cmd is read command, void *data will hold the return
 *        data
 *	  If this function failed, return value -1
 * 	  this function whould use >64 byte stack space.
 *
 */
int submitSafte(U32 cmd, void *data, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	int 	ret;
	int	len;
	char	msg[64] = {0};	/* tmp buf to hold msg */
	Enclosure_Conf *pec = (Enclosure_Conf *)msg;

	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] == NULL)
		return(-1);

	ret = -1;
	memset(msg, 0, 64);

	switch (cmd) {
	case GET_ENC_CONF:
		ret = Safte_GetConf(pec, cardno);
		if(ret >= 0) 
		{
			memcpy((char *)data, (char *)pec, sizeof(Enclosure_Conf));
			return(0);
		} else
			return(-1);	

	case GET_ENC_STAT:
		ret = Safte_GetEnclosureStatus(msg, cardno);
		if(ret > 0)
		{
			memcpy((char *)data, msg, ret);
			return(0);
		} else
			return(-1);

	case GET_ENC_USAGE:
		return(-1);

	case GET_DEV_INSERT:
		return(-1);

	case GET_DEV_STAT:
		ret = Safte_GetDevstat(msg, cardno);
		if(ret > 0) 
		{
			memcpy((char *)data, msg, ret);
			return(0);
		} else
			return(-1);

	case SET_DEV_STAT:
		len = (psc->sstat.devslot_num * sizeof(SetDeviceStat1));
		memcpy((char *)msg, (char *)data+1, len);
		ret = Safte_SetDevStat(msg, cardno);
		return(0);

	case SET_BUS_ID:
		return(-1);	/* No support yet */

	case SET_DEV_OP:
		memcpy((char *)msg, (char *)data + 1, sizeof(SetSlotOp));
		ret = Safte_SetDevOp(msg, cardno);
		if(ret < 0)
		    return(-1);
		else
		    return(0);

	case SET_FAN_SPEED:
		return(-1);	/* No support yet */

	case SET_POWER_STAT:
		return(-1);	/* No support */

	case SET_GLOB_CMD:
		memcpy((char *)msg, (char *)data +1, sizeof(SetGlobalcmd));
		ret = Safte_SetGlobalCmd(msg, cardno);
		if(ret < 0)
		    return(-1);
		else
		    return(0);

	default:
		break;
	}

	return(-1);

}



/*
 *
 */
static int init_saftebox(pSafte_Context_t psc, int cardno)
{
	int i,boxno,sbid,ret;
	HdAryInfo_t hai[MAXCHANNEL] = {{0}};
	unsigned char status;
	
	if(psc == NULL)
		return(-1);
	
	psc->sstat.gs.AudibleAlarm = 0;		/* not support alarm */
	psc->sstat.gs.GlobalFail = 1;	
	psc->sstat.gs.GlobalWarn = 1;
	psc->sstat.gs.EnclosurePower = 0;	/* no support turn on/off power by app */
	psc->sstat.gs.CoolFail = 0;		/* not support fan/cooling failure */
	psc->sstat.gs.PowerFail = 1;		/* indicate power failure */
	psc->sstat.gs.DrvFail = 1;		/* indicate driver failure */
	psc->sstat.gs.DrvWarn = 1;		/* indicate drive warning */
	psc->sstat.gs.ArrayFail = 1;		/* indicate array failure */
	psc->sstat.gs.ArrayWarn = 1;		/* indicate array warning */
	psc->sstat.gs.EnclosureLock = 0;	/* not support to indicate door lock */

	ret = Init_SBOX(cardno, psc->regbaddr, &psc->lstat);

	if(!ret)	return(-1);
	
	psc->sstat.devslot_num = psc->lstat.boxsum;
	
	/* Get device status from FW by OSD (Scnch payload) */
	ret = getChStatusFromFW(MAXCHANNEL, hai, cardno);
	
	/* Try to get each sbox infomation */
	for(i = 0, boxno = 0; i < MAXCHANNEL && boxno < psc->lstat.boxsum; i++)
	{
		/* Try to get SMBUS ID of box connected to this channel */
		if((sbid = psc->lstat.ChIdmapSbId[i]) != 0xff)
		{
			boxno++;
		} 
		else
		{
			continue;
		}

		/* Check power status */
		if(SBox_ReadPowerStatus((U8)sbid, psc->regbaddr)  == PowerON)
		{
			*(unsigned char *)(&psc->lstat.ep) |= bit(i);
		} else {
			*(unsigned char *)(&psc->lstat.ep) &= (~bit(i));
		}
		
		psc->sstat.powsup_num++;
		
		/* check sensor */
		if (SBox_ReadWinBondId(psc->regbaddr, (U8)sbid) == WINBONDW83782DID)
		{
			Sbox_Connect_2ndBus((U8)sbid, psc->regbaddr);
			
			/* Check Fan status */
			psc->lstat.FanSpeed1[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)sbid, (U8)0, (U8)0);
			if(psc->lstat.FanSpeed1[i] > 0)
			{
				psc->sstat.fan_num++;
			}
			if((psc->lstat.BoxType[sbid] & 0x0f) == 0x09)
			{	/* 4100 */
				psc->lstat.FanSpeed2[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)sbid, (U8)1, (U8)0);
				if(psc->lstat.FanSpeed2[i] > 0)
				{
					psc->sstat.fan_num++;
				}
			}

			/* Check tempature */
			psc->lstat.temp1[i] = Sbox_ReadWinBondTemp1(psc->regbaddr, (U8)sbid, (U8)0);
			if(psc->lstat.temp1[i] > 0)
			{
				psc->sstat.tempsen_num++;
			}
			if((psc->lstat.BoxType[sbid] & 0x0f) == 0x09)
			{	/* 4100 */
				psc->lstat.temp2[i] = Sbox_ReadWinBondTempX(psc->regbaddr, (U8)sbid, (U8)1, (U8)0);
				if(psc->lstat.temp2[i] > 0)
				{
					psc->sstat.tempsen_num++;
				}
			}
			
			Sbox_Disconnect_2ndBus((U8)sbid, psc->regbaddr);
		}
		
		/* check whether HD exist on each channel */
		if(ret > 0)
		{
			*(U8 *)&psc->lstat.has[i].ds = hai[i].has; /* save hd array status */
			psc->lstat.has[i].aryid = hai[i].aryid;
			status = hai[i].has;
			if(status == NODRIVE)
			{
				psc->lstat.ArrayStatLed[i] = LED_RED;	/* led0 dark */
				psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */
			} else if(status == NOERROR && hai[i].aryid == FREEDISK) {
				psc->lstat.ArrayStatLed[i] = LED_GREEN;	/* led0 Green */
				psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */
			} else if(status == NOERROR && hai[i].aryid < (unsigned char)MAXCHANNEL) {
				psc->lstat.ArrayStatLed[i] = LED_GREEN;	/* led0 Green */
				psc->lstat.ArrayActiveLed[i] = LEDLIGHT; /* led1 light */
			} else if(status == DEVFAULT || status == INFAILEDARRAY) {
				psc->lstat.ArrayStatLed[i] = LED_RED;	/* led0 red */
				psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */				
			} else if(status == DEVREBUILD || status == INCRITARRAY) {
				psc->lstat.ArrayStatLed[i] = LED_ORANGE;	/* led0 Amber */
				psc->lstat.ArrayActiveLed[i] = LEDLIGHT; /* led1 light */				
			} else {
				psc->lstat.ArrayStatLed[i] = LED_RED;	/* led0 dark */
				psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */
			}
				
			SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
			SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);
		} else {
			*(U8 *)&psc->lstat.has[i].ds = 0;
			psc->lstat.has[i].ds.NoDrive = 1;	/* No HD exist */
			psc->lstat.ArrayStatLed[i] = LED_RED;	/* default led */
			psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led0/led1 dark */
				
			SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
			SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);			
		}
	}

	return(0);

}


/*
 *	Get information of all enclosure connected with smbus
 *	Input:
 *		Enclosure_Conf *pec: a pointer of buffer to hold enclosure information
 *	Output:
 *		return 0 if success
 */
static int Safte_GetConf(Enclosure_Conf *pec, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	unsigned char *ptr1;
	int i, k, boxno;

	/* walk through all enclosure detected */
	for(i = 0, boxno = 0; 
	    i < MAXCHANNEL && boxno < psc->sstat.devslot_num; i++)
	{
	    k = psc->lstat.ChIdmapSbId[i];
	    if(k < 0)	/* no enclosure exist */
		continue;
		
	    boxno++;

	    /* Check power status */
	    ptr1 = (unsigned char *)(&psc->lstat.ep);
	    if(SBox_ReadPowerStatus((unsigned char)k, psc->regbaddr) == PowerON)
	    { 
		*ptr1 |= bit(i);
	    } else {	/* Power Off */
		*ptr1 &= (~bit(i));
	    }
	    pec->powsup_num++;
	    
	    /* check sensor */
	    if (SBox_ReadWinBondId(psc->regbaddr, (U8)k) == WINBONDW83782DID)
	    {	
		/* Check Fan status */
		psc->lstat.FanSpeed1[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)k, (U8)0, (U8)1);
		if(psc->lstat.FanSpeed1[i] >= 0)
			pec->fan_num++;
		if((psc->lstat.BoxType[k] & 0x0f) == 0x09)
		{	/* 4100 */
			psc->lstat.FanSpeed2[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)k, (U8)1, (U8)1);
			if(psc->lstat.FanSpeed2[i] >= 0)
				pec->fan_num++;
		}

		/* Check tempature */
		psc->lstat.temp1[i] = Sbox_ReadWinBondTemp1(psc->regbaddr, (U8)k, (U8)1);
		if(psc->lstat.temp1[i] >= 0)
			pec->tempsen_num++;
		if((psc->lstat.BoxType[k] & 0x0f) == 0x09)
		{	/* 4100 */
			psc->lstat.temp2[i] = Sbox_ReadWinBondTempX(psc->regbaddr, (U8)k, (U8)1, (U8)1);
			if(psc->lstat.temp2[i] >= 0)
				pec->tempsen_num++;
		}
	    }
	}

	psc->sstat.powsup_num = pec->powsup_num;
	psc->sstat.fan_num = pec->fan_num;
	psc->sstat.tempsen_num = pec->tempsen_num;

	pec->devslot_num = psc->sstat.devslot_num;

	pec->doorlock = DOOR_UNKNOW;
	pec->speaker = SP_OFF;
	pec->vendspec_num = 0;

	psc->sstat.doorlock = pec->doorlock;
	psc->sstat.speaker = pec->speaker;
	
	return(0);
}

/*
 * Get the status of all enclosure
 *	Input:
 *		a pointer of buffer that will hold all enclosure status.
 *	Output:
 *		return a positive value of the data bytes if success	
 */
static int Safte_GetEnclosureStatus(char *pbuf, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	int i, k, boxno;
	int fannum, pownum, tempnum;
	int powoffset, devoffset, tempoffset, tempflagoffset;
	int ret;
	U8 *ptr;	/* ???### */

	fannum = psc->sstat.fan_num;
	pownum = psc->sstat.powsup_num;
	tempnum = psc->sstat.tempsen_num;

	powoffset = fannum;	/* power supply items following fan items */
	devoffset = fannum + pownum;	/* device slot following power supply */
	tempoffset = devoffset + psc->sstat.devslot_num + 2; /* 2 is for doorlock and speaker */
	tempflagoffset = tempoffset + psc->sstat.tempsen_num * sizeof(U8); /* ???### */

	ptr = (U8 *)&pbuf[tempoffset]; /* ???### */
	
	pbuf[psc->sstat.devslot_num+fannum+pownum] = psc->sstat.doorlock;

	pbuf[psc->sstat.devslot_num+fannum+pownum+1] = psc->sstat.speaker;

	/* walk through all enclosure detected */
	for(i = 0, boxno = 0;
	    i < MAXCHANNEL && boxno < psc->sstat.devslot_num; i++)
	{
	    k = psc->lstat.ChIdmapSbId[i];
	    if(k < 0)	/* no enclosure exist */
		continue;
		
	    pbuf[devoffset + boxno] = (char)k;

	    /* Check power status */
	    if((ret=SBox_ReadPowerStatus((U8)k, psc->regbaddr)) == PowerON)
	    {
		pbuf[powoffset + psc->sstat.powsup_num - pownum] = PowerON;
	    } else if(ret == PowerOFF) {
		pbuf[psc->sstat.fan_num+psc->sstat.powsup_num-pownum] = PowerOFF;
	    } else {
		pbuf[psc->sstat.fan_num+psc->sstat.powsup_num-pownum] = POW_UNKNOWN;
	    }
	    pownum--;
	    
	    if (SBox_ReadWinBondId(psc->regbaddr, (U8)k) == WINBONDW83782DID)
	    {
		/* check fan status */
		psc->lstat.FanSpeed1[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)k, (U8)0, (U8)1);
		if(psc->lstat.FanSpeed1[i] > FAN_STOP_RPM) 
		{
			pbuf[psc->sstat.fan_num-fannum] = FAN_OP;
		} else {
			pbuf[psc->sstat.fan_num-fannum] = FAN_MALFUN;
		}
		fannum--;

		if((psc->lstat.BoxType[k] & 0x0f) == 0x09)
		{	/* 4100 */
			psc->lstat.FanSpeed2[i] = Sbox_ReadWinBondFanSpeed(psc->regbaddr, (U8)k, (U8)1, (U8)1);
			if(psc->lstat.FanSpeed2[i] > FAN_STOP_RPM) 
			{
				pbuf[psc->sstat.fan_num-fannum] = FAN_OP;
			} else {
				pbuf[psc->sstat.fan_num-fannum] = FAN_MALFUN;
			}
			fannum--;
		}
		
		/* Check tempature */
		psc->lstat.temp1[i] = Sbox_ReadWinBondTemp1(psc->regbaddr, (U8)k, (U8)1);
		if((psc->lstat.temp1[i] > 0) && 
			psc->lstat.temp1[i] < OVERTEMP_LIMIT)
		{

			*ptr++ = CELSIUS2FAHRENHEIT(psc->lstat.temp1[i]);
			pbuf[tempflagoffset+1] &= (~bit((psc->sstat.tempsen_num-tempnum)));
		} else { /* an abnormal temperature */
			*ptr++ = CELSIUS2FAHRENHEIT(psc->lstat.temp1[i]);
			pbuf[tempflagoffset+1] |= bit((psc->sstat.tempsen_num-tempnum));
		}
		tempnum--;

		if((psc->lstat.BoxType[k] & 0x0f) == 0x09)
		{	/* 4100 */
			psc->lstat.temp2[i] = Sbox_ReadWinBondTempX(psc->regbaddr, (U8)k, (U8)1, (U8)1);
			if((psc->lstat.temp2[i] > 0) &&
				psc->lstat.temp2[i] < OVERTEMP_LIMIT)
			{
				*ptr++=CELSIUS2FAHRENHEIT(psc->lstat.temp2[i]);
				pbuf[tempflagoffset+1] &= (~bit((psc->sstat.tempsen_num-tempnum)));
			} else { /* an abnormal temperature */
				*ptr++=CELSIUS2FAHRENHEIT(psc->lstat.temp2[i]);
				pbuf[tempflagoffset+1] |= bit((psc->sstat.tempsen_num-tempnum));
			}
			tempnum--;
		}
	    }

	    boxno++;
	}

	for(; fannum > 0; fannum--)
		pbuf[psc->sstat.fan_num - fannum] = FAN_MALFUN;

	for(; tempnum > 0; tempnum--)
	{
		*(U8 *)&pbuf[tempoffset + (psc->sstat.tempsen_num-tempnum) * sizeof(U8)] = 0; /* ???### */
		pbuf[tempflagoffset+1] |= bit((psc->sstat.tempsen_num-tempnum));
	}

	return(tempflagoffset+2);
}

/*
 * Get status of all devices inserted in inclosure
 *	Input:
 *		char *pbuf: a pointer of buffer that will hold the status
 *	Output:
 *		return a positive value of data bytes
 */
static int Safte_GetDevstat(char *pbuf, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	GetDeviceStat1	*pds1 = (GetDeviceStat1 *)pbuf;
	DeviceStat_t *ps = NULL;
	int i;
	unsigned char boxno, sbid;
	
	for(i = 0, boxno = 0; 
	    i < MAXCHANNEL && boxno < psc->sstat.devslot_num; i++)
	{
		sbid = psc->lstat.ChIdmapSbId[i];
		if(sbid == 0xff)
			continue;

		ps = (DeviceStat_t *)&psc->lstat.has[i].ds;
		if(ps->NoDrive == 0)
		{
			/* copy drive status */
			pds1[boxno].sds = psc->lstat.has[i].ds; 
			pds1[boxno].DevInserted = 1;	/* device inserted */
			pds1[boxno].ReadyInsRev = 1;	/* ready */
			if(ps->devfault == 0 && ps->InFailedArray == 0)
			{
				pds1[boxno].PreparedOp=1;/* activated */
			} else {
				pds1[boxno].PreparedOp=0;/*can't be activated*/
			}				
		} else {
			pds1[boxno].sds.NoDrive = 1;  /* so we set no device */
			pds1[boxno].DevInserted = 0;  /* no device inserted */
			pds1[boxno].ReadyInsRev = 0;  /* not ready */
			pds1[boxno].PreparedOp = 0;   /* not activated */
		}
		boxno++;
	}

	return(psc->sstat.devslot_num * sizeof(GetDeviceStat1));
}

/*
 *	Set all devices and enclosure status
 *	Input:
 *		char *pbuf:	The status buffer
 *	Output:
 *		return 0 if success, or return -1
 */
static int Safte_SetDevStat(char *pbuf, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	SetDeviceStat1 *psds = (SetDeviceStat1 *)pbuf;
	int i, boxno;
	unsigned char led, gled;
	unsigned char sbid;

	for(i = 0, boxno = 0; 
	    i < MAXCHANNEL && boxno < psc->sstat.devslot_num; i++)
	{
		sbid = psc->lstat.ChIdmapSbId[i];
		if(sbid == 0xff)
			continue;

		led = 0; gled = 0;

		if(psds[boxno].sds.noerror)		     /* no error */
		{
			led = LED_GREEN;		     /* led 0green */
			gled = psc->lstat.ArrayActiveLed[i]; /* keep old led1 */
		} 
		else if(psds[boxno].sds.devfault) 
		{					/* Device Faulty */
			led = LED_RED;			/* led0 red */
			gled = LEDDARK;			/* led1 dark */
		} 
		else if(psds[boxno].sds.devrebuild) 
		{ 					/* Device Rebuilding */
			led = LED_ORANGE;		/* led0 range */
			gled = LEDFASTFLASH;		/* led1 slow flash */
		}
		else if(psds[boxno].sds.InFailedArray)
		{				 	/* In Failed Array */
			led = LED_RED;			/* led0 Red */
			gled = LEDDARK;			/* led1 dark */
		}
		else if(psds[boxno].sds.InCritArray)
		{ 					/* In Critical Array */
			led = LED_ORANGE;		/* led0 orange */
			gled = psc->lstat.ArrayActiveLed[i]; /* keep old led1 */
		}
		else if(psds[boxno].sds.ParityCheck)
		{				 	/* Parity Check */
			led = LED_ORANGE;		/* led0 orange */
			gled = psc->lstat.ArrayActiveLed[i]; /* keep old led1 */
		}
		else if(psds[boxno].sds.PredictedFault)
		{					/* Predicted Fault */
			led = LED_ORANGE;		/* led0 orange */
			gled = psc->lstat.ArrayActiveLed[i]; /* keep old led1 */
		}
		else if(psds[boxno].sds.NoDrive)
		{ 				/* No Drive */
			led = LED_RED;		/* Red */
			gled = LEDDARK;		/* led0 light, led1 dark */
		}
		
		psc->lstat.ArrayStatLed[i] = led;	/* save array status led */
		psc->lstat.ArrayActiveLed[i] = gled;	/* save global led control */
		psc->lstat.has[i].ds = psds[boxno].sds;	/* save drive status */
		
		SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
		SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);	
		boxno++;
		
	}

	return(0);
}

/*
 * Perform Slot Operation
 */

static int Safte_SetDevOp(char *pbuf, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	SetSlotOp	*psso = (SetSlotOp *)pbuf;
	int bId, i;

	if(psso->ChannelNo > psc->sstat.devslot_num)
		return(-1);

	bId = psc->lstat.ChIdmapSbId[psso->ChannelNo];
	if(psso->PrepOp || psso->Identify ||psso->PrepInsRev)
	{
		if(psso->Identify)
		{
			for(i = 0; i < 3; i++)
			{
				/* Turn on yellow */
				Sbox_simple((U8)bId, (U8)6, psc->regbaddr);

				/* Set Yellow led flash fast  blink */
				Sbox_write((U8)bId, (U8)9, (U8)0x3, psc->regbaddr);

				/* Turn off yellow led */
				Sbox_simple((U8)bId, (U8)7, psc->regbaddr);
			}
			
			/* restore led status */
			Sbox_write((U8)bId, (U8)9, (U8)0x00, psc->regbaddr);
			
			SBox_SetStatusLED((U8)bId, 
				psc->lstat.ArrayActiveLed[psso->ChannelNo],
				psc->regbaddr);
			
			SBox_SetFlashLED((U8)bId, 
				psc->lstat.ArrayStatLed[psso->ChannelNo], 
				psc->regbaddr);
							
		}
		return(0);
	}
	
	return(-1);
}

/*
 * Set Global Command
 *	Input Parameter
 *		char *pbuf: the pointer of a buf including global stat will be set

 *	Output
  *		return value 0 if success, or return -1
 *
 */
static int Safte_SetGlobalCmd(char *pbuf, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	unsigned short *pd, *ps;

	pd = (unsigned short *)&psc->sstat.gs;
	ps = (unsigned short *)&pbuf[1];
	*pd = *ps;

	return(0);
}

/*
 * As market strange request:
 *	 set status led to be orange, set flash led to dark after
 * OS shutdown
 * It's only called by cml_notifier_func() during OS shutdown
 */
int SetLedInShutdown(int chanid, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	unsigned char sbid;
	
	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] == NULL)
		return(-1);
		
	/* Try to get SMBUS ID of box connected to this channel */
	if((sbid = psc->lstat.ChIdmapSbId[chanid]) == 0xff)
	{
		return(-1);
	} 

	SBox_SetStatusLED((U8)sbid, LED_OFF, psc->regbaddr);
	SBox_SetFlashLED((U8)sbid, LEDDARK, psc->regbaddr);						
	return(0);
}

/*
 * The function is used to set drive status led of enclosure on each channel
 * as status of array the drive be assigned to
 *
 * Input:
 *	int aryid:	The array Id present which array is processed
 *
 *	int ArrayStat:
 *		Array status: ARRAY_NORMAL(0x00) for normal, 
 *			      ARRAY_WARNING(0x01) for warning,
 *			      ARRAY_ERROR(0x10) for error
 *	
 *	int HDnum:	How manay hard disk the array has
 *
 *	unsigned char HDList:
 *		It is an 32-bit mapping, each 4 bits present each channel; 
 *		bits [4n, 4n+1, 4n+2,4n+3] isn't 0x0e or 0x0f to indicate 
 * 		the corresponding channel is assigned to this array.
 *		those [4n,...,4n+3] is 0x0f to indicate the disk on this channel 
 *		isn't assigned to this array if disk exist on this channel or 
 *		no hard disk exist on this channel
 *		[4n...4n+3] is equal to 0x0e means the hard disk is failed or offline
 *
 * Ouput:
 *	If success, return value 0; or return -1;
 *
 */

int	SetArrayStatLed(int aryid, int ArrayStat, int HDnum, U32 HDList, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	int i, j, hdid, boxno;
	unsigned char sbid;
	
	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] == NULL)
		return(-1);
		
	if(ArrayStat != ARRAY_NORMAL &&	ArrayStat != ARRAY_WARNING && ArrayStat != ARRAY_ERROR)
		return(-1);
			
	for(i = 0, j=0, boxno = 0; 
		i < MAXCHANNEL && boxno < psc->lstat.boxsum; i++)
	{

		/* Try to get SMBUS ID of box connected to this channel */
		if((sbid = psc->lstat.ChIdmapSbId[i]) != 0xff)
		{
			boxno++;
		} 
		else
		{
			continue;
		}

		hdid = ((HDList >> (i*4)) & 0x0f); /* 4 bits for each channel */

		/*
		 * Notice: FREEDISK is a special array id !!!
		 */
		 
		/* Check whether the hd is assigned to this array before */
		if(psc->lstat.has[i].aryid == aryid)
		{
			if(hdid == 0x0f)	/* the hd has been removed from this array */
			{
				psc->lstat.has[i].aryid = FREEDISK;
				if(psc->lstat.has[i].ds.NoDrive == 0 || 
					psc->lstat.ArrayActiveLed[i] != LEDDARK)
				{	/* turn off both leds */
					*(U8 *)&psc->lstat.has[i].ds = 0;
					psc->lstat.has[i].ds.NoDrive = 1;
					psc->lstat.ArrayActiveLed[i] = LEDDARK;
					psc->lstat.ArrayStatLed[i] = LED_OFF;
					SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
					SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);						
				}
				continue;
			}
		} else {
			if(hdid != 0x0f)
				psc->lstat.has[i].aryid = aryid;
			else
				continue;
		}
			
		j++;

		if(hdid == 0x0e )	/* bad disk */
		{
			if(psc->lstat.has[i].ds.devfault == 1 &&
				psc->lstat.ArrayStatLed[i] == LED_RED &&
				psc->lstat.ArrayActiveLed[i] == LEDDARK &&
				psc->lstat.statledinited)
			{
				continue;	/* optimize performance */
			}
				
			*(U8 *)&psc->lstat.has[i].ds = 0;
			psc->lstat.has[i].ds.devfault = 1;
			psc->lstat.has[i].ds.NoDrive = 1; /* ???### */
			psc->lstat.ArrayStatLed[i] = LED_RED;	/* led0 red */
			psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */
			SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
			SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);			
			continue;
		}
		
		switch (ArrayStat)
		{
		case	ARRAY_NORMAL:
			/* optimize: if previous status is same, nothing to do */
			if(psc->lstat.has[i].ds.noerror == 1 &&
			   psc->lstat.ArrayStatLed[i] == LED_GREEN &&
			   ((aryid != FREEDISK &&
				psc->lstat.ArrayActiveLed[i] != LEDDARK) ||
			    (aryid == FREEDISK &&
				psc->lstat.ArrayActiveLed[i] == LEDDARK)) &&
			   hdid != 0x0f /* noexist */ &&
			   hdid != 0x0e /* bad hd */ &&
			   psc->lstat.statledinited)
			{
				continue;
			}
				
			*(U8 *)&psc->lstat.has[i].ds = 0;/*clear drive status*/
			psc->lstat.has[i].ds.noerror = 1;
			psc->lstat.ArrayStatLed[i] = LED_GREEN;	/* green */
			if(aryid != FREEDISK) /* No Free or Spare disk */
			{
				if(psc->lstat.ArrayActiveLed[i] == LEDDARK)
					psc->lstat.ArrayActiveLed[i] = LEDLIGHT;
			} else {	/* Spare or Free Disk */
				psc->lstat.ArrayActiveLed[i] = LEDDARK;
			}
			break;
				
		case	ARRAY_WARNING:
			/*
			 * optimize performance:
			 *	if previous status is same, nothing to do
			 */
			if(psc->lstat.has[i].ds.InCritArray == 1 &&
			   psc->lstat.ArrayStatLed[i] == LED_ORANGE &&
			   psc->lstat.ArrayActiveLed[i] != LEDDARK  &&
			   hdid != 0x0f /* noexist */ &&
			   hdid != 0x0e /* bad disk */&&
			   psc->lstat.statledinited)
			{
				continue;
			}
				
			*(U8 *)&psc->lstat.has[i].ds = 0;/*clear drive status*/
			psc->lstat.has[i].ds.InCritArray = 1;
			psc->lstat.ArrayStatLed[i] = LED_ORANGE;/*led0 amber*/
			if(psc->lstat.ArrayActiveLed[i] == LEDDARK)
				psc->lstat.ArrayActiveLed[i] = LEDLIGHT; /* led1 light */
			break;
				
		case	ARRAY_ERROR:
			/*
			 * optimize performance:
			 *	if previous status is same, nothing to do
			 */
			if(psc->lstat.has[i].ds.InFailedArray == 1 &&
			   psc->lstat.ArrayStatLed[i] == LED_RED &&
			   psc->lstat.ArrayActiveLed[i] == LEDDARK &&
			   hdid != 0x0f /* no exist */ &&
			   psc->lstat.statledinited)
			{
				continue;
			}
				
			*(U8 *)&psc->lstat.has[i].ds = 0;
			psc->lstat.has[i].ds.NoDrive = 1; /* ???### */
			psc->lstat.has[i].ds.InFailedArray = 1;
			psc->lstat.ArrayStatLed[i] = LED_RED;	/* led0 red */
			psc->lstat.ArrayActiveLed[i] = LEDDARK;	/* led1 dark */
			break;
						
		default:
			continue;		
		}
		
		SBox_SetStatusLED((U8)sbid, psc->lstat.ArrayStatLed[i], psc->regbaddr);
		SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);
	}

	psc->lstat.statledinited = 1;

	if(j != HDnum)
		return(-1);

	return(0);		

}
 


/*
 * The function is used to set drive status led of enclosure on each channel
 * as status of array the drive be assigned to
 *
 * Input:
 *	unsigned char HDList:
 *		It is an 8-bit mapping; bit n is set one to indicate 
 * 		the disk on the corresponding channel is accessed, or
 *		means the disk on the channel is idle if exist
 *	
 * Ouput:
 *	If success, return value 0; or return -1;
 *
 */

int	SetDiskWorkLed(unsigned char HDList, int cardno)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	int i, boxno, sbid;

	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] == NULL)
		return(-1);
		

	for(i = 0, boxno = 0; i < MAXCHANNEL && boxno < psc->lstat.boxsum; i++)
	{

		/* Try to get SMBUS ID of box connected to this channel */
		if((sbid = psc->lstat.ChIdmapSbId[i]) != 0xff)
		{
			boxno++;
		} 
		else
		{
			continue;
		}

		if(HDList & bit(i))	/* Working: flash */
		{
			if(psc->lstat.ArrayActiveLed[i] == LEDFASTFLASH &&
			   psc->lstat.workledinited) /* optimize performance */
			{
				continue;
			} else {
				psc->lstat.ArrayActiveLed[i] = LEDFASTFLASH;
			}
		} else {	/* not working */
			if(psc->lstat.has[i].ds.NoDrive || 
				psc->lstat.has[i].ds.devfault ||
				psc->lstat.has[i].ds.InFailedArray ||
				psc->lstat.has[i].aryid == FREEDISK ||
				psc->lstat.ArrayStatLed[i] == LED_OFF ||
				psc->lstat.ArrayStatLed[i] == LED_RED)
			{ 	/* no drive or disk error: dark*/
				if(psc->lstat.ArrayActiveLed[i] == LEDDARK &&
				   psc->lstat.workledinited)  /* optimize */
				{
					continue;
				}
				else
				{
					psc->lstat.ArrayActiveLed[i] = LEDDARK;
				}
			} else { /* Disk exist, light not flash */
				if(psc->lstat.ArrayActiveLed[i] == LEDLIGHT &&
				   psc->lstat.workledinited)
				{
					continue;
				}
				else
				{
					psc->lstat.ArrayActiveLed[i] = LEDLIGHT;
				}
			}
		}
		
		SBox_SetFlashLED((U8)sbid, psc->lstat.ArrayActiveLed[i], psc->regbaddr);
	}

	psc->lstat.workledinited = 1;

	return(0);			
}



/*
 * The function will be called to set the enclosure led to be plugin or 
 * plug-out/error status;
 * The function should be called when "Plug-In" or "Set Disk Down" Event happened
 * And it will call SetArrayStatLed() internal
 *	Input:
 *		char cardno:	The carmel card's no.
 *		char chid:	The No of the channel the disk located at
 *		char status:	Present the disk's status
 *				0 --- Plug-Out/Error;
 *				1 --- Plug-In
 *
 *	Out:
 *		if success, return 0, or return -1
 *
 */
int SetDiskPlugInOutLed(int cardno, char chid, char status)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];
	U32	hdlist;
	int i, hdnum;
	char aryid;

	if(cardno < 0 || cardno >= MAX_CARDS || pSafteContext[cardno] == NULL)
		return(-1);
	
	if(chid < 0 || chid >= MAXCHANNEL || (status != DISKPLUGOUT && status != DISKPLUGIN))
		return(-1);
		
	hdnum = 0;
	hdlist = 0;
	i = cml_check_raidfeature(cardno, 1);
	if(i == 0)
		aryid = FREEDISK;
	else
		aryid = chid;

	for(i = 0; i < MAXCHANNEL; i++)
	{
		if(i == chid)
		{ /* this disk */
			hdnum++;
			if(status == DISKPLUGOUT)	/* Plug-out/Error */
			{
				hdlist |= (0x0e << (4*i));
			} else {			/* Plug-in */
				hdlist |= (i << (4*i));
			}
		} else	if(aryid == psc->lstat.has[i].aryid && 
			   psc->lstat.has[i].ds.NoDrive == 0) 
		{	/* Just for another free/spare disk ie. aryid is FREEDISK */
			hdnum++;
			hdlist |= (i << (4*i));
		} else {
			hdlist |= (0x0f << (4*i));
		}
	}
	
	if(status == DISKPLUGOUT)
	{
		SetArrayStatLed(aryid, ARRAY_ERROR, hdnum, hdlist, cardno);
	} else {
		SetArrayStatLed(aryid, ARRAY_NORMAL, hdnum, hdlist, cardno);
	}
	
	return(0);	
}

unsigned char GetChanMapSmbus(int cardno, int channo)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];

	if(psc == NULL)
		return(0xff);

	if(psc->lstat.ChIdmapSbId[channo] != 0xff)
		return(psc->lstat.ChIdmapSbId[channo]);

	return(0xff);
}

void SetChIdmapSbId(int cardno, int chan, unsigned char sbid)
{
	pSafte_Context_t psc = (pSafte_Context_t)pSafteContext[cardno];

	if(psc == NULL)
		return;

	if(psc->lstat.ChIdmapSbId[chan] == 0xff)
		psc->lstat.ChIdmapSbId[chan] = sbid;

	return;
}
