/*+M*************************************************************************
 * Promise SATAII S150 SX8 device driver for Linux.
 *
 * Copyright (c) 2001  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2001 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *-M************************************************************************
 */
 
/*
 * Description   : smbus.c - The codes to control smbus
 */

#if defined (__KERNEL__) | (__FreeBSD__)	
    #define PACKED __attribute__((packed))
#else  
    #define PACKED
#endif

#if defined (__KERNEL__)
#include <stdarg.h>
#include <asm/io.h>
#include <asm/string.h>
#include <linux/delay.h>
#endif

#include "smbus.h"
#include "tools.h"

static void SMBus_Delay(unsigned long);
static void SMBus_Start(U8 *);
static void SMBus_Stop(U8 *);
static U8 SMBus_Get_SDA(U8 *);
static void SMBus_Set_SDA0(U8 *);
static void SMBus_Set_SDA1(U8 *);
static void WriteByte(U8, U8 *);
static U8 ReadByte(U8 *);
static U8 WaitforAck(U8 *);

static U8 bAx;

static void SMBus_Delay(unsigned long delay)
{
	udelay(delay);
}

static void SMBus_Start(U8 *baddr)
{
    bAx = (SM_SCL | SM_SDA);
    cml_out8(baddr, SMST, bAx);

    /* delay 1 ms to reactive*/
    SMBus_Delay(5);
    
    bAx &= (~SM_SDA);    
    SMBus_Delay(5);
    cml_out8(baddr, SMST, bAx);

    bAx &= (~SM_SCL);    
    SMBus_Delay(5);
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(5);
}

static void SMBus_Stop(U8 *baddr)
{
    bAx &= ~(SM_SDA | SM_SCL);
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(5);
    
    bAx |= SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(5);
    
    bAx |= SM_SDA;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(20);	
}

static void SMBus_Set_SDA0(U8 *baddr)
{
    bAx &= ~(SM_SDA);
    
    bAx &= ~(SM_SCL);
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);

    bAx |= SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(8);

    bAx &= ~SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);
}

static void SMBus_Set_SDA1(U8 *baddr)
{
    bAx |= SM_SDA;
    bAx &= ~SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);

    bAx |= SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(8);

    bAx &= ~SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);
}

static U8 SMBus_Get_SDA(U8 *baddr)
{
    U8 ret;
 
    bAx &= ~SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);
    
    bAx |= SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);
    
    ret = cml_in8(baddr, SMST);
    ret = ((ret & SM_SDA) ? 1: 0);
    SMBus_Delay(4);
     
    bAx &= ~SM_SCL;
    cml_out8(baddr, SMST, bAx);
    SMBus_Delay(4);
    
    return ret;
}

static void WriteByte(U8 data, U8 *baddr)
{
    int ii;
    
    /* MSB written first*/
    for(ii=0; ii<8; ii++)
    {
        if (data & (1 << (7 - ii)))
            SMBus_Set_SDA1(baddr);
        else
            SMBus_Set_SDA0(baddr);
    }

    /* 
     * Raise Data 
     */ 
    bAx |= SM_SDA;
    cml_out8(baddr, SMST, bAx);
    
    /* Set output disable. */
    bAx = cml_in8(baddr, SMST);
  
}

static U8 ReadByte(U8 *baddr)
{
 /* MSB read first, shift left */
    U8 ret = 0, ii;

    /* Set SDA=1 because SDA is open-drain signal */    
    bAx |= SM_SDA;
    cml_out8(baddr, SMST, bAx);
    
    for(ii=0; ii<8; ii++)
    {
        ret = ret << 1;
        ret |= SMBus_Get_SDA(baddr);
    }

    return(ret);
}


/* 
 * PLS notice the slave address of SMBus is 7 bits only. The 0 bit bus must be 0
 */

static U8 WaitforAck(U8 *baddr)
{
    U8 Ack;
    /* U32 count = 0; */
    
    /* Set output disable.*/
    if ((Ack = SMBus_Get_SDA(baddr)) == 0x01)
    {
   	return(0x00);
    }
    else 
    	return(0x01);
}

int SMBus_init(U8 *baddr)
{
    cml_out8(baddr, SMIM, 0);

    return(0);
}

U16 SMBus_ReadWord(U8 SlaveAddr, U8 Command, U8 *baddr)
{
    U8 Address, LowByte, HighByte;
    
    U8 Ack = 0x00;
    U16 ret = 0xFFFF;
      
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_WRITE | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(ret);
    	
    WriteByte(Command, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(ret);
        
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_READ | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(ret);

    LowByte = ReadByte(baddr);
    /* Send Ack */
    SMBus_Set_SDA0(baddr);

    HighByte = ReadByte(baddr);
    /* Send NonAck */
    SMBus_Set_SDA1(baddr);
    
    SMBus_Stop(baddr);
    
    ret = (U16)HighByte << 8;
    ret |= (U16) LowByte;
    return(ret);
}

/*
 * This function support the write word. 	 
 */
void SMBus_WriteWord(U8 SlaveAddr, U8 Command, U16 Data, U8 *baddr)
{
    U8 Address, LowByte, HighByte;
    /* U16 ret =0xFFFF; */
    U8 Ack = 0x00;
      
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_WRITE | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    {
	SMBus_Stop(baddr);
    	return;
    }
    	
    WriteByte(Command, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    {
	SMBus_Stop(baddr);
    	return;
    }
    
    
    LowByte = (U8)Data;
    WriteByte(LowByte, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    {
	SMBus_Stop(baddr);
    	return;
    }

    HighByte = (U8) Data >> 0x08;
    WriteByte(HighByte, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    {
	SMBus_Stop(baddr);
    	return;
    }

    SMBus_Stop(baddr);
}

/*
	This function support the block read. In SMBus protocol, 
	A block reador write/read is allowed to xfer a max. of 32 byte
	 
*/
U8 SMBus_ReadBlock(U8 SlaveAddr, U8 Command, PU8 pBuf, U8 bLen, U8 *baddr)
{
    U8 Address;
    U8 bLength;
    U8 Ack = 0x00;

    SMBus_Start(baddr);
    Address = (SM_ADDRESS_WRITE | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);

    WriteByte(Command, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);
    
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_READ | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);

    /* get xfer length */
    bLength = ReadByte(baddr);
    /* Send Ack */
    SMBus_Set_SDA0(baddr);
    
    if (bLength > bLen) 
    	return(0);
    else
    {
    	U8 idx;
    	for (idx = 0; idx < bLength; idx ++)
    	{

    	    *(pBuf++) = ReadByte(baddr);
    	    
    	     /* Send Non Ack for the last one data */
    	     if (idx == bLength -1 )
    	     	SMBus_Set_SDA1(baddr);
    	     else 
    		SMBus_Set_SDA0(baddr);
    	}
    }
    return(1);
}


int SMBus_Simple(U8 SlaveAddr, U8 *baddr)
{
    U8 Address; 
    U8 Ack = 0x00;
    
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_WRITE | SlaveAddr);
    
    WriteByte(Address, baddr);
    
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);
    
    SMBus_Stop(baddr);
    return(1);
}

U8 SMBus_ReadOneByte(U8 SlaveAddr, U8 *baddr)
{
    U8 Address; 
    U8 Ack = 0x00;
    U8 ret = 0xFF;
      
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_READ | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(ret);
    
    ret = ReadByte(baddr);
    /* Send NoAck, use smbus_set_sda1 or delay a clock. */
    SMBus_Set_SDA1(baddr);
    
    SMBus_Stop(baddr);
    
    return(ret);
}

int SMBus_WriteOneByte(U8 SlaveAddr, U8 Data, U8 *baddr)
{
    U8 Address; /*, LowByte, HighByte; */
    /* U16 ret =0xFFFF; */
    U8 Ack = 0x00;
      
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_WRITE | SlaveAddr);
    WriteByte(Address, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);
    
    WriteByte(Data, baddr);
    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    	return(0);
    
    SMBus_Stop(baddr);

    return(1);
}

U16 SMBus_WinbondReadTwoBytesFromPresetPointerLocation(U8 SlaveAddr, U8 *baddr)
{
    U8 Address; 
    U8 Ack = 0x00;
    U16 ret = 0xFFFF;
      
    SMBus_Start(baddr);
    Address = (SM_ADDRESS_READ | SlaveAddr);
    WriteByte(Address, baddr);

    Ack = WaitforAck(baddr);
    if (Ack == 0x00)
    {
    	return(ret);
    }

    ret = ReadByte(baddr);
    ret <<= 8;

    SMBus_Set_SDA0(baddr);

    ret |= ReadByte(baddr);

    SMBus_Set_SDA1(baddr);

    SMBus_Stop(baddr);
 
    return(ret);
}

int	SMBus_WriteBytes(U8 *baddr, U8 *pbuf, int len)
{
	U8 Ack = 0x00;
	int i;
	
	if(pbuf == (U8 *)0 || len > 32)
		return(0);
		
	SMBus_Start(baddr);
    	
    	for(i = 0; i < len; i++)
    	{
		WriteByte(pbuf[i], baddr);
		
		Ack = WaitforAck(baddr);
		if (Ack == 0x00) 
		{
			SMBus_Stop(baddr);  			
			return(0);
    		}
    	}
    
	SMBus_Stop(baddr);  
	return(len); 	
}
